# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from copy import deepcopy
from optimizn.combinatorial.branch_and_bound import BnBProblem


class TravelingSalesmanProblem(BnBProblem):
    def __init__(self, params):
        self.input_graph = params['input_graph']
        # sort all distance values, for computing lower bounds
        self.sorted_dists = []
        for i in range(self.input_graph.dists.shape[0]):
            for j in range(0, i):
                self.sorted_dists.append(self.input_graph.dists[i, j])
        self.sorted_dists.sort()
        super().__init__(params)

    def _get_closest_city(self, city, visited):
        # get the unvisited city closest to the one provided
        min_city = None
        min_dist = float('inf')
        dists = self.input_graph.dists[city]
        for i in range(len(dists)):
            if i != city and i not in visited and dists[i] < min_dist:
                min_city = i
                min_dist = dists[i]
        return min_city

    def _complete_path(self, path):
        '''
        This function performs the nearest-neighbor approximation algorithm
        for the traveling salesman problem, which is presented in the following
        source.

        Source:

        (1)
        Title: An Analysis of Several Heuristics for the Traveling Salesman
        Problem
        Authors: Daniel J. Rosenkrantz, Richard E. Stearns, and Philip M.
        Lewis II
        Journal: SIAM Journal on Computing
        Volume: 6
        Number: 3
        DOI: 10.1137/0206041
        URL: https://www.researchgate.net/publication/220616869_An_Analysis_of_Several_Heuristics_for_the_Traveling_Salesman_Problem
        Date published: September 1977
        Date accessed: 2021
        '''

        # complete the path greedily, iteratively adding the unvisited city
        # closest to the last city in the accumulated path
        visited = set(path)
        complete_path = deepcopy(path)
        while len(complete_path) != self.input_graph.dists.shape[0]:
            if len(complete_path) == 0:
                next_city = 0
            else:
                last_city_idx = 0 if len(complete_path) == 0 else\
                    complete_path[-1]
                next_city = self._get_closest_city(last_city_idx, visited)
            visited.add(next_city)
            complete_path.append(next_city)
        return complete_path

    def get_candidate(self):
        # greedily assemble a path from scratch
        return self._complete_path([])
    
    def get_root(self):
        # return empty path
        return []

    def complete_solution(self, sol):
        # greedily complete the path using the remaining/unvisited cities
        return self._complete_path(sol)

    def cost(self, sol):
        # sum of distances between adjacent cities in path, and from last
        # city to first city in path
        path_cost = 0
        for i in range(self.input_graph.num_cities - 1):
            path_cost += self.input_graph.dists[sol[i], sol[i + 1]]
        path_cost += self.input_graph.dists[
            sol[self.input_graph.num_cities - 1], sol[0]]
        return path_cost

    def lbound(self, sol):
        # sum of distances between cities in path and smallest distances
        # to account for remaining cities and start city
        num_cities_in_path = len(sol)
        lb_path_cost = 0
        for i in range(num_cities_in_path - 1):
            lb_path_cost += self.input_graph.dists[sol[i], sol[i + 1]]
        if num_cities_in_path == self.input_graph.num_cities:
            lb_path_cost += self.input_graph.dists[
                sol[num_cities_in_path - 1], sol[0]]
        elif num_cities_in_path == 0:
            lb_path_cost += sum(self.sorted_dists[
                :self.input_graph.num_cities])
        else:
            lb_path_cost += sum(self.sorted_dists[
                :self.input_graph.num_cities - num_cities_in_path + 1])
        return lb_path_cost

    def is_complete(self, sol):
        # check that all cities covered once, path length is equal to the
        # number of cities
        check_all_cities_covered = set(sol) == set(
            range(self.input_graph.num_cities))
        check_cities_covered_once = len(sol) == len(set(sol))
        check_path_length = len(sol) == self.input_graph.num_cities
        return (check_path_length and check_cities_covered_once and
                check_all_cities_covered)

    def is_feasible(self, sol):
        # check that covered cities are only covered once and path length is
        # less than or equal to the number of cities
        check_cities_covered_once = len(sol) == len(set(sol))
        check_path_length = len(sol) <= self.input_graph.num_cities
        return check_cities_covered_once and check_path_length

    def branch(self, sol):
        # build the path by creating a new solution for each uncovered city,
        # where the uncovered city is the next city in the path
        if len(sol) == self.input_graph.num_cities:
            return []
        visited = set(sol)
        new_sols = []
        for new_city in range(self.input_graph.dists.shape[0]):
            if new_city not in visited:
                new_sols.append(sol + [new_city])
        return new_sols
